/*********************************************************************
*               SEGGER MICROCONTROLLER GmbH & Co KG                  *
*       Solutions for real time microcontroller applications         *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2008  SEGGER Microcontroller GmbH & Co KG         *
*                                                                    *
*       www.segger.com     Support: support@segger.com               *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system for microcontrollers      *
*                                                                    *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS version: 3.62a                                         *
*                                                                    *
**********************************************************************

----------------------------------------------------------------------
File    : RTOSInit_AVR32AP7000.c               (for AP7000 CPU)
Purpose : Initializes and handles the hardware for the OS as far
          as required by the OS.
--------  END-OF-HEADER  ---------------------------------------------
*/

#include "RTOS.H"

#include "prep.h"
#include "intc.h"

/*********************************************************************
*
*       Configuration
*
*********************************************************************/

/*********************************************************************
*
*       Clock frequency settings
*/
#define   OS_FSYS       (150000000L)

#ifndef   OS_TICK_FREQ
  #define OS_TICK_FREQ  (1000)
#endif

#ifndef   OS_PCLK_TIMER                /* Clock used for embOS timer */
  #define OS_PCLK_TIMER (OS_FSYS)
#endif

#ifndef   OS_PCLK_UART                 /* Peripheral clock for UART  */
  #define OS_PCLK_UART  (OS_FSYS/2)
#endif

/*********************************************************************
*
*       UART settings for OSView
*       If you do not want (or can not due to hardware limitations)
*       to dedicate a UART to OSView, please define it to be -1
*       Currently the standard code enables UART 0 per default
*/
#ifndef   OS_UART
  #define OS_UART (1)
#endif

#ifndef   OS_BAUDRATE
  #define OS_BAUDRATE (38400)
#endif

/*********************************************************************
*
*		Check configuration
*/
#define BAUDRATEREG (OS_PCLK_UART/(OS_BAUDRATE*16))
#if BAUDRATEREG > 0xFFFF
#error "Cannot set up UART for the given OS_PCLK_UART and OS_BAUDRATE values."
#endif

/****** End of configuration settings *******************************/
#define OS_UART_USED (OS_UART == 1)

/*********************************************************************
*
*       Local defines (sfrs used in RTOSInit.c)
*
**********************************************************************
*/
#define AVR32B_PM_BASE_ADDR  0xFFF00000
#define PLL_MCCTRL           (*(volatile OS_U32*)(AVR32B_PM_BASE_ADDR + 0x00)) // Main Clock Control MCCTRL Read/Write 0x0
#define PLL_CKSEL            (*(volatile OS_U32*)(AVR32B_PM_BASE_ADDR + 0x04)) // Clock Select  Read/Write 0x0
#define PLL_CPUMASK          (*(volatile OS_U32*)(AVR32B_PM_BASE_ADDR + 0x08)) // CPU Clock Mask CPUMASK Read/Write Impl. defined
#define PLL_HSBMASK          (*(volatile OS_U32*)(AVR32B_PM_BASE_ADDR + 0x0C)) // HSB Clock Mask HSBMASK Read/Write Impl. defined
#define PLL_PBAMASK          (*(volatile OS_U32*)(AVR32B_PM_BASE_ADDR + 0x10)) // PBA Clock Mask PBAMASK Read/Write Impl. defined
#define PLL_PBBMASKR         (*(volatile OS_U32*)(AVR32B_PM_BASE_ADDR + 0x14)) // PBB Clock Mask PBBMASK Read/Write Impl. defined
#define PLL_PLL0             (*(volatile OS_U32*)(AVR32B_PM_BASE_ADDR + 0x20)) // PLL0 Control PLL0 Read/Write 0x0
#define PLL_PLL1             (*(volatile OS_U32*)(AVR32B_PM_BASE_ADDR + 0x24)) // PLL1 Control PLL1 Read/Write 0x0
#define PLL_ISR              (*(volatile OS_U32*)(AVR32B_PM_BASE_ADDR + 0x4C)) // Interrupt Status
#define PLL0_IS_SOURCE       2
#define PLLCOUNT             20
#define PLLMUL               15  
#define PLLDIV               2
#define PLLOPT               4  
#define PLLOSC               0   // Use Oscillator 0 as source for PLL
#define PLLEN                1   // PLL is enabled  
#define LOCK0                1
#define _COUNT               0x108
#define _COMPARE             0x10c

#define SMC0_SETUP (*(volatile OS_U32*)(0xFFF03400))
#define SMC0_PULSE (*(volatile OS_U32*)(0xFFF03404))
#define SMC0_CYCLE (*(volatile OS_U32*)(0xFFF03408))
#define SMC0_MODE  (*(volatile OS_U32*)(0xFFF0340C))

typedef void OS_ISR_HANDLER(void);

#define OS_TIMER_RELOAD (OS_PCLK_TIMER/OS_TICK_FREQ)
#if (OS_TIMER_RELOAD >= 0x100000000)
  #error "Timer 1 can not be used, please check configuration"
#endif

/*********************************************************************
*
*       Prototypes
*
**********************************************************************
*/
void OS_InitInterrupts(void);
void OS_InstallISRHandler (OS_ISR_HANDLER* pISRHandler, int ISRIndex, int ISRPrio);

/*********************************************************************
*
*       Externals
*
**********************************************************************
*/
extern const unsigned int ipr_val[AVR32_INTC_NUM_INT_LEVELS];  // Values to store in the interrupt priority registers for the four interrupt priority levels

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/
// Creates a table of interrupt line handlers per interrupt group in order to optimize RAM space.
// Each line handler table contains a set of pointers to interrupt handlers.
#define DECL_INT_LINE_HANDLER_TABLE(GRP, unused) \
static volatile __int_handler _int_line_handler_table_##GRP[Max(AVR32_INTC_NUM_IRQS_PER_GRP##GRP, 1)];
MREPEAT(AVR32_INTC_NUM_INT_GRPS, DECL_INT_LINE_HANDLER_TABLE, ~);
#undef DECL_INT_LINE_HANDLER_TABLE

// Table containing for each interrupt group the number of interrupt request
// lines and a pointer to the table of interrupt line handlers.
static const struct
{
  unsigned int num_irqs;
  volatile __int_handler *_int_line_handler_table;
} _int_handler_table[AVR32_INTC_NUM_INT_GRPS] =
{
#define INSERT_INT_LINE_HANDLER_TABLE(GRP, unused) \
  {AVR32_INTC_NUM_IRQS_PER_GRP##GRP, _int_line_handler_table_##GRP},
  MREPEAT(AVR32_INTC_NUM_INT_GRPS, INSERT_INT_LINE_HANDLER_TABLE, ~)
#undef INSERT_INT_LINE_HANDLER_TABLE
};


/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/ 

/*********************************************************************
*
*       _InitSMC()
*/
static void _InitSMC(void) {  
  //
  // Initialize external memory timing parameter, before enabling PLL, ensure at least one hold cycle
  //
  SMC0_CYCLE = 0
             | (0x09 << 16)    // NRD_CYCLE length (has to be equal or larger than NCS_RD_SETUP + NCS_RD_PULSE_WIDTH eg. NRD_SETUP + NRD_PULSE_WIDTH
             | (0x09 << 0)     // NWR_CYCLE length (has to be equal or larger than NCS_WR_SETUP + NCS_WR_PULSE_WIDTH eg. NWE_SETUP + NWE_PULSE_WIDTH
             ;
  SMC0_PULSE = 0
             | (0x06 << 24)    // NCS_RD_PULSE_WIDTH
             | (0x06 << 16)    // NRD_PULSE_WIDTH
             | (0x06 << 8)     // NCS_WR_PULSE_WIDTH
             | (0x04 << 0)     // NWE_PULSE_WIDTH
             ; 
  SMC0_SETUP = 0
	         | (0x00 << 24)    // NCS_RD_SETUP = 0 , not required for connected FLASH
	         | (0x00 << 16)    // NRD_SETUP    = 0 , not required for connected FLASH
	         | (0x00 << 8)     // NCS_WR_SETUP = 0 
	         | (0x03 << 0)     // NWE_SETUP    = 3
	         ;
  //
  // Setup access mode for Flash, Set TDF optimization, 1 data float cycle, nRD controlled read mode
  // Writing the mode register will update the timing parameter which were set before
  //
  //SMC0_MODE  = (SMC0_MODE & ~(0x1F << 16)) | (1 << 20) | (1 << 16) | (1 << 0);
}

/*********************************************************************
*
*       _InitPLL()
*/
static void _InitPLL(void) {  
  PLL_CKSEL = 0
            | (0uL << 7)  | (0uL << 0)       // CPU_DIV = 0; CPUSEL = 0; => CPU clock = main clock
            | (1uL << 15) | (0uL << 8)       // HSB_DIV = 1; HSBSEL = 0; => HSB clock = main clock / 2
            | (1uL << 23) | (0uL << 16)      // PBA_DIV = 1; PBASEL = 0; => PBA clock = main clock / 2
            | (1uL << 31) | (1uL << 24)      // PBB_DIV = 1; PBBSEL = 1; => PBB clock = main clock / 4
            ;   
  //
  // Initialize PLL
  //
  PLL_PLL0  = (PLLCOUNT       << 24)     // Number of slow clocks
            | ((PLLMUL - 1)   << 16)     // PLL Multiplier
            | ((PLLDIV - 1)   << 8)      // PLL Divider
            | (PLLOPT         << 2)      // PLL Options (always set to 100b)
            | (PLLOSC         << 1)      // Oscillator 0 is source for PLL 
            | (PLLEN          << 0);     // Enable PLL  
  while ((PLL_ISR & LOCK0) == 0);  
  PLL_MCCTRL = PLL0_IS_SOURCE;    // Set PLL0 as clock source for main clock
}


/*********************************************************************
*
*       _OS_ISR_TickHandler()
*/
static void _OS_ISR_TickHandler(void) __attribute__((interrupt)); 
static void _OS_ISR_TickHandler(void) {
  OS_I32 t, t1;
  OS_EnterInterrupt();
  t  = OS_GetCompare();
  do {
    t +=  OS_TIMER_RELOAD;
    OS_TICK_Handle();
    t1 = OS_GetCount() + 1000;
  } while ((t - t1) < 0);
  OS_SetCompare(t);
  OS_LeaveInterrupt();
}

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/

/*********************************************************************
*
*       __low_level_init()
*/
int __low_level_init(void);
int __low_level_init(void) {
  return 1;
}


/*********************************************************************
*
*       OS_InitHW()
*
*       Initialize the hardware (timer) required for the OS to run.
*       May be modified, if an other timer should be used
*/
void OS_InitHW(void) {
  int count;
  OS_IncDI();
  //
  // Initialize external memory timing parameter, before enabling PLL, ensure at least one hold cycle
  //
  _InitSMC();  
  //
  // Before enabling PLL, set clock divider fCPU = 2*fHSB, 2*fPBA, 4*fPBB
  //
  _InitPLL();
  //
  // Initialize OS timer
  //
  count = OS_GetCount();
  OS_SetCompare(count + OS_TIMER_RELOAD);
  //
  // Set auto interrupt address and priority 0 for int0 group (count/compare timer)
  //
  OS_InitInterrupts();
  OS_InstallISRHandler(_OS_ISR_TickHandler, 0, 0);
  //
  // Finally initialize UART for embOSView
  //
  OS_COM_Init();
  OS_DecRI();
}

/*********************************************************************
*
*       Idle loop  (OS_Idle)
*
*       Please note:
*       This is basically the "core" of the idle loop.
*       This core loop can be changed, but:
*       The idle loop does not have a stack of its own, therefore no
*       functionality should be implemented that relies on the stack
*       to be preserved. However, a simple program loop can be programmed
*       (like toggeling an output or incrementing a counter)
*/
void OS_Idle(void) {     // Idle loop: No task is ready to exec
  while (1) {
#if DEBUG == 0
// Instruction to put CPU in halt mode to save power
#endif
  }
}

/*********************************************************************
*
*       Get time [cycles]
*
*       This routine is required for high
*       resolution time maesurement functions.
*       It returns the system time in timer clock cycles.
*/
OS_U32 OS_GetTime_Cycles(void) {
  OS_U32 time;
  int    count;
  int    compare;
  time  = OS_Time;
  compare = OS_GetCompare();
  count   = compare - OS_GetCount();
  if (count < 0) {  // missed a counter interrupt, adjust time
    time++;
    count = 0 - count;
  } else {
    count = OS_TIMER_RELOAD - count;
  }
  return (OS_PCLK_TIMER/OS_TICK_FREQ) * time + count;
}

/*********************************************************************
*
*       OS_ConvertCycles2us
*
*       Convert Cycles into micro seconds.
*
*       If your clock frequency is not a multiple of 1 MHz,
*       you may have to modify this routine in order to get proper
*       diagonstics.
*
*       This routine is required for profiling or high resolution time
*       measurement only. It does not affect operation of the OS.
*/
OS_U32 OS_ConvertCycles2us(OS_U32 Cycles) {
  return Cycles/(OS_PCLK_TIMER/1000000);
}

/*********************************************************************
*
*       Communication for OSView via UART (optional)
*
**********************************************************************
*/

#if OS_UART_USED

#define USART1_BASE_ADDR     0xFFE01000

#define USART_CR             (*(volatile OS_U32*) (USART1_BASE_ADDR + 0x00))
#define USART_MR             (*(volatile OS_U32*) (USART1_BASE_ADDR + 0x04))
#define USART_IER            (*(volatile OS_U32*) (USART1_BASE_ADDR + 0x08))
#define USART_IDR            (*(volatile OS_U32*) (USART1_BASE_ADDR + 0x0C))
#define USART_IMR            (*(volatile OS_U32*) (USART1_BASE_ADDR + 0x10))
#define USART_CSR            (*(volatile OS_U32*) (USART1_BASE_ADDR + 0x14))
#define USART_RHR            (*(volatile OS_U32*) (USART1_BASE_ADDR + 0x18))
#define USART_THR            (*(volatile OS_U32*) (USART1_BASE_ADDR + 0x1C))
#define USART_BRGR           (*(volatile OS_U32*) (USART1_BASE_ADDR + 0x20))
#define USART_RTOR           (*(volatile OS_U32*) (USART1_BASE_ADDR + 0x24))
#define USART_TTGR           (*(volatile OS_U32*) (USART1_BASE_ADDR + 0x28))

#define AVR32B_GPIO_PIOA_BASE_ADDR       0xFFE02800
#define GPIO_PIOA_ASR         (*(volatile OS_U32*) (AVR32B_GPIO_PIOA_BASE_ADDR + 0x70))  // Peripheral A Select Register
#define GIPO_PIOA_PDR         (*(volatile unsigned int*) (AVR32B_GPIO_PIOA_BASE_ADDR + 0x04))

static char _TxChar;

static void _UartHandler(void) __attribute__((interrupt));
static void _UartHandler(void) {
  OS_U32 c;
  OS_EnterInterrupt();
  if ((USART_CSR & 0x00000001) != 0) {
    c = USART_RHR;
    OS_OnRx(c);
  }
  if ((USART_CSR & 0x00000002) != 0) {
    USART_THR = _TxChar;
    if (OS_OnTx()) {              // No more characters to send ?
      USART_IDR = 0x00000002;     // interrupt disabled for: TXEMPTY
    }
  }
  OS_LeaveInterrupt();
}

/*********************************************************************
*
*       OS_COM_Send1()
*       Never call this function directly from your application
*/
void OS_COM_Send1(OS_U8 c) {
  _TxChar = c;
  USART_IER = 0x00000002;       // interrupt enabled for: TXRDY
}

/*********************************************************************
*
*       OS_COM_Init()
*       Initialize UART for OSView
*/
void OS_COM_Init(void) {            /* Initialize UART, enable UART interrupts */
  GPIO_PIOA_ASR = (1 << 17) | (1 << 18); // PA17 and PA18 peripheral Mode A
  GIPO_PIOA_PDR = (1 << 17) | (1 << 18); // Disable PIO control, pin controlled by the periphera
  USART_IDR   = 0xFFFFFFFF;              // disable all uart interrupts
  USART_MR    = 0x00000000;              // reset mode register
  USART_RTOR  = 0x00000000;              // reset receive timout register
  USART_TTGR  = 0x00000000;              // reset transmitter timeguard register
  USART_CR    = 0x0000610C;              // reset usart, disable dtr and rts
  USART_BRGR  = BAUDRATEREG;             // set baudrate to selected baud rate
  USART_MR    = 0x000008C0;              // set mode: Normal, MCK, 8 data bits, async mode, even parity, 1 stop bit, normal mode, OVER=8
  USART_IER   = 0x00000001;              // interrupt enabled for: RXRDY
  USART_CR   |= 0x00000050;              // enable rx and tx
  OS_InstallISRHandler(_UartHandler, 32*7, 0);
}

#else  /* selected UART not supported, using dummies */

void OS_COM_Init(void) {}
void OS_COM_Send1(OS_U8 c) {
  OS_USEPARA(c);           /* avoid compiler warning */
  OS_COM_ClearTxActive();  /* let the OS know that Tx is not busy */
}

#endif /*  OS_UART_USED  */

/****** Final check of configuration ********************************/

#ifndef OS_UART_USED
  #error "OS_UART_USED has to be defined"
#endif

/*********************************************************************
*
*       OS interrupt handler and ISR specific functions
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_GetIRQHandler
*
*       Detect reason for IRQ and returns correspondig service routine addr.
*       OS_irq_handler is called from intx() function found in exception.s
*/
OS_ISR_HANDLER* OS_GetIRQHandler(int ISRPrio) {
  unsigned int int_grp, int_req;
  int_grp = (&AVR32_INTC.icr0)[ISRPrio];	    // Get group number
  int_req = AVR32_INTC.irr[int_grp];          // Get Irq number in group
  return (int_req) ? _int_handler_table[int_grp]._int_line_handler_table[32 - clz(int_req) - 1] : NULL;
}

/*********************************************************************
*
*       OS_Unhandled_Interrupt
*/
void OS_Unhandled_Interrupt(void)
{
  OS_Error(OS_ERR_ISR_NO_HANDLER);         // Call OS_Error() to inform about missing interrupt handler
}

/*********************************************************************
*
*       OS_ARM_InstallISRHandler
*/
void OS_InstallISRHandler (OS_ISR_HANDLER* pISRHandler, int ISRIndex, int ISRPrio) {
  unsigned int int_grp;
  int_grp = ISRIndex / AVR32_INTC_MAX_NUM_IRQS_PER_GRP;  // Calculate group number for given ISRIndex
  _int_handler_table[int_grp]._int_line_handler_table[ISRIndex % AVR32_INTC_MAX_NUM_IRQS_PER_GRP] = pISRHandler;  // Set handler addr in int table
  AVR32_INTC.ipr[int_grp] = ipr_val[ISRPrio];            // Set Interrupt priority in IPR register of selected group
}

/*********************************************************************
*
*       OS_InitInterrupts
*
*       Initializes the interrupt table
*       Each entry is filled with the default interrupt handler 
*       OS_Unhandled_Interrupt()
*/
void OS_InitInterrupts(void) {
  int int_grp, int_req;

  // For all interrupt groups,
  for (int_grp = 0; int_grp < AVR32_INTC_NUM_INT_GRPS; int_grp++)
  {
    // For all interrupt request lines of each group,
    for (int_req = 0; int_req < _int_handler_table[int_grp].num_irqs; int_req++)
    {
      // Assign _unhandled_interrupt as default interrupt handler.
      _int_handler_table[int_grp]._int_line_handler_table[int_req] = &OS_Unhandled_Interrupt;
    }

    // Set the interrupt group priority register to its default value.
    // By default, all interrupt groups are linked to the interrupt priority
    // level 0 and to the interrupt vector _int0.
    AVR32_INTC.ipr[int_grp] = ipr_val[INT0];
  }
}

/*****  EOF  ********************************************************/
